@@ -73,11 +73,12 @@ union sighandler __sighandler[NSIG] =
 
 /* The wrapper around sigaction.  Install our own signal handler
    around the signal. */
-int __sigaction(int sig, const struct sigaction * act,
-              struct sigaction * oact)
+int __pthread_sigaction(int sig, const struct sigaction * act,
+			struct sigaction * oact)
 {
   struct sigaction newact;
   struct sigaction *newactp;
+  __sighandler_t old = SIG_DFL;
 
   if (sig == __pthread_sig_restart ||
       sig == __pthread_sig_cancel ||
@@ -86,6 +87,8 @@ int __sigaction(int sig, const struct sigaction * act,
       __set_errno (EINVAL);
       return -1;
     }
+  if (sig > 0 && sig < NSIG)
+    old = (__sighandler_t) __sighandler[sig].old;
   if (act)
     {
       newact = *act;
@@ -96,21 +99,27 @@ int __sigaction(int sig, const struct sigaction * act,
 	    newact.sa_handler = (__sighandler_t) __pthread_sighandler_rt;
 	  else
 	    newact.sa_handler = (__sighandler_t) __pthread_sighandler;
+	  if (old == SIG_IGN || old == SIG_DFL || old == SIG_ERR)
+	    __sighandler[sig].old = (arch_sighandler_t) act->sa_handler;
 	}
       newactp = &newact;
     }
   else
     newactp = NULL;
   if (__libc_sigaction(sig, newactp, oact) == -1)
-    return -1;
+    {
+      if (act)
+	__sighandler[sig].old = (arch_sighandler_t) old;
+      return -1;
+    }
   if (sig > 0 && sig < NSIG)
     {
       if (oact != NULL
 	  /* We may have inherited SIG_IGN from the parent, so return the
 	     kernel's idea of the signal handler the first time
 	     through.  */
-	  && (__sighandler_t) __sighandler[sig].old != SIG_ERR)
-	oact->sa_handler = (__sighandler_t) __sighandler[sig].old;
+	  && old != SIG_ERR)
+	oact->sa_handler = old;
       if (act)
 	/* For the assignment it does not matter whether it's a normal
 	   or real-time signal.  */
@@ -118,10 +127,13 @@ int __sigaction(int sig, const struct sigaction * act,
     }
   return 0;
 }
-strong_alias(__sigaction, sigaction)
+#ifdef SHARED
+strong_alias(__pthread_sigaction, __sigaction)
+strong_alias(__pthread_sigaction, sigaction)
+#endif
 
 /* sigwait -- synchronously wait for a signal */
-int sigwait(const sigset_t * set, int * sig)
+int __pthread_sigwait(const sigset_t * set, int * sig)
 {
   volatile pthread_descr self = thread_self();
   sigset_t mask;
@@ -173,10 +185,13 @@ int sigwait(const sigset_t * set, int * sig)
   *sig = THREAD_GETMEM(self, p_signal);
   return 0;
 }
+#ifdef SHARED
+strong_alias (__pthread_sigwait, sigwait)
+#endif
 
 /* Redefine raise() to send signal to calling thread only,
    as per POSIX 1003.1c */
-int raise (int sig)
+int __pthread_raise (int sig)
 {
   int retcode = pthread_kill(pthread_self(), sig);
   if (retcode == 0)
@@ -186,3 +201,9 @@ int raise (int sig)
     return -1;
   }
 }
+#ifdef SHARED
+strong_alias (__pthread_raise, raise)
+#endif
+
+/* This files handles cancellation internally.  */
+LIBC_CANCEL_HANDLED ();
